#define _GNU_SOURCE
#include <err.h>
#include <stdbool.h>
#include <stdint.h>
#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/prctl.h>
#include <sys/mman.h>
#include <sys/eventfd.h>
#include <sys/syscall.h>
#include <linux/futex.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static int fd;
static char *mapping;

int main(void) {
  fd = SYSCHK(syscall(__NR_memfd_create, "memfd", 0));
  mapping = SYSCHK(mmap((void*)0x200000, 0x200000, PROT_READ|PROT_WRITE, MAP_SHARED|MAP_FIXED_NOREPLACE, fd, 0));

  /* first allocate a single page before hugepage eligibility */
  SYSCHK(ftruncate(fd, 0x1000));
  mapping[0] = 'a';

  SYSCHK(ftruncate(fd, 0x200000));
  for (int i=0; i<512; i++)
    *(volatile char *)(mapping+i*0x1000) = 'a';

  int evfd = SYSCHK(eventfd(0, 0));

  SYSCHK(madvise(mapping, 0x200000, MADV_HUGEPAGE));
  pid_t child = SYSCHK(fork());
  if (child == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    if (getppid() == 1) exit(0);

    /* fork() skips shared VMAs to be faster, repopulate page tables */
    for (int i=0; i<512; i++)
      *(volatile char *)(mapping+i*0x1000) = 'a';

    int pm_fd = SYSCHK(open("/proc/self/pagemap", O_RDONLY));
    for (unsigned long iter=0; true; iter++) {
      uint64_t pm_entry;
      off_t pm_off = (unsigned long)mapping / 0x1000 * sizeof(pm_entry);
      assert(SYSCHK(pread(pm_fd, &pm_entry, sizeof(pm_entry), pm_off)) == sizeof(pm_entry));
      if ((pm_entry & (1ULL << 63)) == 0) {
        printf("child: pagemap entry gone in iter %lu\n", iter);
        break;
      }
    }

    usleep(500000);
    *(volatile char *)mapping;
    SYSCHK(eventfd_write(evfd, 1));

    while (1) pause();
  }

  sleep(1); /* let child do its prep */
  int pm_fd = SYSCHK(open("/proc/self/pagemap", O_RDONLY));
  //SYSCHK(madvise(mapping, 0x200000, MADV_HUGEPAGE));

  for (unsigned long iter=0; true; iter++) {
    uint64_t pm_entry;
    off_t pm_off = (unsigned long)mapping / 0x1000 * sizeof(pm_entry);
    assert(SYSCHK(pread(pm_fd, &pm_entry, sizeof(pm_entry), pm_off)) == sizeof(pm_entry));
    if ((pm_entry & (1ULL << 63)) == 0) {
      printf("pagemap entry gone in iter %lu\n", iter);
      *(volatile char *)mapping;
      break;
    }
  }

  eventfd_t dummy_ev;
  SYSCHK(eventfd_read(evfd, &dummy_ev));

  for (unsigned long iter=0; true; iter++) {
    uint64_t pm_entry;
    off_t pm_off = (unsigned long)mapping / 0x1000 * sizeof(pm_entry);
    assert(SYSCHK(pread(pm_fd, &pm_entry, sizeof(pm_entry), pm_off)) == sizeof(pm_entry));
    if ((pm_entry & (1ULL << 63)) == 0) {
      printf("pagemap entry gone in iter %lu ???\n", iter);
      usleep(500000);
      *(volatile char *)mapping;
    }
    if (pm_entry & (1ULL << 56)) {
      printf("pagemap poll stop at iter %lu\n", iter);
      break;
    }
  }

  SYSCHK(fallocate(fd, FALLOC_FL_PUNCH_HOLE|FALLOC_FL_KEEP_SIZE, 0, 0x200000));
  sleep(5);
  return 0;
}
